from concurrent.futures import CancelledError
from types import TracebackType
from typing import Any, Generator, Generic, Tuple, TypeVar

import trio


class InvalidState(Exception):
    pass


TResult = TypeVar("TResult")


class Future(Generic[TResult]):
    _error: Tuple[BaseException, TracebackType]
    _result: TResult

    _cancelled: bool
    _did_error: bool

    _done: trio.Event

    def __init__(self) -> None:
        self._cancelled = False
        self._did_error = False
        self._done = trio.Event()

    def cancelled(self) -> bool:
        return self._cancelled

    def done(self) -> bool:
        return self._done.is_set()

    def result(self) -> TResult:
        if self.cancelled():
            raise CancelledError
        elif self.done():
            if self._did_error:
                exc_value, exc_tb = self._error
                raise exc_value.with_traceback(exc_tb)
            else:
                return self._result
        else:
            raise InvalidState()

    def set_result(self, result: TResult) -> None:
        self._result = result
        self._done.set()

    def set_exception(self, exc_value: Exception, exc_tb: TracebackType) -> None:
        self._error = exc_value, exc_tb
        self._done.set()
        self._did_error = True

    def cancel(self) -> bool:
        if self.done():
            return False
        self._done.set()
        self._cancelled = True
        return True

    def exception(self) -> BaseException:
        if self.cancelled():
            raise CancelledError
        elif self.done() and self._did_error:
            return self._error[0]
        else:
            raise InvalidState()

    async def _await(self) -> TResult:
        await self._done.wait()
        return self.result()

    def __await__(self) -> Generator[Any, None, TResult]:
        return self._await().__await__()



class LahjaError(Exception):
    """
    Base class for all lahja errors
    """

    pass


class BindError(LahjaError):
    """
    Raise when an attempt was made to bind an event that is already bound.
    """


class LifecycleError(LahjaError):
    """
    Raised when attempting to violate the lifecycle of an endpoint such as
    starting an already started endpoint or starting an endpoint that has
    already stopped.
    """

    pass


class ConnectionAttemptRejected(LahjaError):
    """
    Raised when an attempt was made to connect to an endpoint that is already connected.
    """

    pass


class UnexpectedResponse(LahjaError):
    """
    Raised when the type of a response did not match the ``expected_response_type``.
    """

    pass


class RemoteDisconnected(LahjaError):
    """
    Raise when a remote disconnects while we attempting to read a message.
    """

    pass



from abc import ABC, abstractmethod
import asyncio
from typing import Any, Awaitable, Callable, Type, cast

import trio

from lahja import AsyncioEndpoint, EndpointAPI, TrioEndpoint
from lahja.base import EventAPI

Driver = Callable[["EngineAPI"], Awaitable[None]]


class EngineAPI(ABC):
    endpoint_class: Type[EndpointAPI]
    Event: Type[EventAPI]

    @abstractmethod
    def run_drivers(self, *drivers: Driver) -> Awaitable[None]:
        """
        Performs the actual *running* of the drivers executing them with in a
        manner appropriate for the individual endpoint implementation.
        """
        ...

    @abstractmethod
    async def run_with_timeout(
        self, coro: Callable[..., Awaitable[Any]], *args: Any, timeout: int
    ) -> None:
        """
        Runs a coroutine with the specifid positional ``args`` with a timeout.
        **must** raise the built-in ``TimeoutError`` when a timeout occurs.
        """
        ...

    @abstractmethod
    async def sleep(self, seconds: float) -> None:
        """
        Sleep for the provide number of seconds in a manner appropriate for the
        individual endpoint implementation.
        """
        ...


class AsyncioEngine(EngineAPI):
    endpoint_class = AsyncioEndpoint
    Event = cast(Type[EventAPI], asyncio.Event)

    async def run_drivers(self, *drivers: Driver) -> None:
        await asyncio.gather(*(driver(self) for driver in drivers))

    async def run_with_timeout(
        self, coro: Callable[..., Awaitable[Any]], *args: Any, timeout: int
    ) -> None:
        try:
            await asyncio.wait_for(coro(*args), timeout=timeout)
        except asyncio.TimeoutError as err:
            raise TimeoutError from err

    async def sleep(self, seconds: float) -> None:
        await asyncio.sleep(seconds)


class TrioEngine(EngineAPI):
    endpoint_class = TrioEndpoint
    Event = cast(Type[EventAPI], trio.Event)

    async def run_drivers(self, *drivers: Driver) -> None:
        async with trio.open_nursery() as nursery:
            for driver in drivers:
                nursery.start_soon(driver, self)

    async def run_with_timeout(
        self, coro: Callable[..., Awaitable[Any]], *args: Any, timeout: int
    ) -> None:
        try:
            with trio.fail_after(timeout):
                await coro(*args)
        except trio.TooSlowError as err:
            raise TimeoutError from err

    async def sleep(self, seconds: float) -> None:
        await trio.sleep(seconds)



import functools
import logging
from typing import Sequence

from lahja.tools.engine import Driver, EngineAPI

from .actions import Action, AsyncAction, SyncAction
from .initializers import Initializer

logger = logging.getLogger("lahja.driver")


class _Driver:
    def __str__(self) -> str:
        initializer_and_actions = " > ".join(
            (str(self.initializer),) + tuple(str(action) for action in self.actions)
        )
        return f"<Driver {initializer_and_actions}>"

    def __init__(self, initializer: Initializer, actions: Sequence[Action]) -> None:
        functools.update_wrapper(self, drive)
        self.initializer = initializer
        self.actions = tuple(actions)

    async def __call__(self, engine: EngineAPI) -> None:
        await drive(engine, self.initializer, self.actions)


def driver(initializer: Initializer, *actions: Action) -> Driver:
    """
    Construct a *Driver*.  Should contain a single *Initializer* followed by a
    variadic number of *Actions*.
    """
    return _Driver(initializer=initializer, actions=actions)


async def drive(
    engine: EngineAPI,
    initializer: Initializer,
    actions: Sequence[Action],
    action_timeout: int = 1,
) -> None:
    """
    Use the provide *Engine* to initialize and drive an endpoint
    implementation.
    """
    logger.debug("STARTING DRIVER")
    async with initializer(engine) as endpoint:
        for idx, action in enumerate(actions):
            logger.debug("RUNNING ACTION[%d]: %s", idx, action)
            if isinstance(action, AsyncAction):
                try:
                    await engine.run_with_timeout(
                        action, engine, endpoint, timeout=action_timeout
                    )
                except TimeoutError as err:
                    raise TimeoutError(
                        f"[{endpoint}] Timeout running action #{idx}:{action}"
                    ) from err
            elif isinstance(action, SyncAction):
                action(engine, endpoint)
            else:
                raise TypeError(f"Unsupported action: {action}")



from abc import ABC, abstractmethod
import asyncio
import itertools
import multiprocessing

import trio

from .engine import AsyncioEngine, Driver, TrioEngine


class RunnerAPI(ABC):
    @abstractmethod
    def __call__(self, *drivers: Driver) -> None:
        ...


class AsyncioRunner(RunnerAPI):
    def __init__(self) -> None:
        self._engine = AsyncioEngine()

    def __call__(self, *drivers: Driver) -> None:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._engine.run_drivers(*drivers))


class TrioRunner(RunnerAPI):
    def __init__(self) -> None:
        self._engine = TrioEngine()

    def __call__(self, *drivers: Driver) -> None:
        trio.run(self._engine.run_drivers, *drivers)


class BaseIsolatedProcessRunner(RunnerAPI):
    def __call__(self, *drivers: Driver) -> None:
        drivers_and_runners = tuple((driver, self.get_runner()) for driver in drivers)
        procs = tuple(
            multiprocessing.Process(target=runner, args=(driver,))
            for driver, runner in drivers_and_runners
        )
        for proc in procs:
            proc.start()
        for proc in procs:
            proc.join()

    @abstractmethod
    def get_runner(self) -> RunnerAPI:
        ...


class IsolatedHomogenousRunner(BaseIsolatedProcessRunner):
    def __init__(self, runner: RunnerAPI):
        self._runner = runner

    def get_runner(self) -> RunnerAPI:
        return self._runner


class IsolatedHeterogenousRunner(BaseIsolatedProcessRunner):
    def __init__(self, *choices: RunnerAPI):
        self._choices = tuple(choices)
        self._choices_iter = itertools.cycle(self._choices)

    def get_runner(self) -> RunnerAPI:
        return next(self._choices_iter)



from typing import NamedTuple, Type

from lahja import BaseEvent, BaseRequestResponseEvent


class RawMeasureEntry(NamedTuple):
    sent_at: float
    received_at: float


class CrunchedMeasureEntry(NamedTuple):
    sent_at: float
    received_at: float
    duration: float


class PerfMeasureEvent(BaseEvent):
    def __init__(self, payload: bytes, index: int, sent_at: float) -> None:
        self.payload = payload
        self.index = index
        self.sent_at = sent_at


class PerfMeasureResponse(BaseEvent):
    pass


class PerfMeasureRequest(BaseRequestResponseEvent[PerfMeasureResponse]):
    def __init__(self, payload: bytes, index: int, sent_at: float) -> None:
        self.payload = payload
        self.index = index
        self.sent_at = sent_at

    @staticmethod
    def expected_response_type() -> Type[PerfMeasureResponse]:
        return PerfMeasureResponse


class ShutdownEvent(BaseEvent):
    pass


class Total(NamedTuple):
    caption: str
    num_total: int
    duration_fastest: float
    duration_slowest: float
    duration_avg: float
    total_aggregated_time: float
    total_duration: float
    first_sent: float
    last_received: float


class TotalRecordedEvent(BaseEvent):
    def __init__(self, total: Total) -> None:
        self.total = total



import time
from typing import List  # noqa: F401

from lahja.tools.benchmark.typing import CrunchedMeasureEntry, RawMeasureEntry, Total


class LocalStatistic:
    def __init__(self) -> None:
        self._entries: List[RawMeasureEntry] = []

    def add(self, entry: RawMeasureEntry) -> None:
        self._entries.append(entry)

    def crunch(self, caption: str) -> Total:
        crunched_entries = []
        for entry in self._entries:
            crunched_entries.append(
                CrunchedMeasureEntry(
                    entry.sent_at, entry.received_at, entry.received_at - entry.sent_at
                )
            )

        fastest = min(n.duration for n in crunched_entries)
        slowest = max(n.duration for n in crunched_entries)
        first_sent = min(n.sent_at for n in crunched_entries)
        last_received = max(n.received_at for n in crunched_entries)
        total_duration = last_received - first_sent
        total_aggregated_time = sum(n.duration for n in crunched_entries)
        avg = total_aggregated_time / len(crunched_entries)

        return Total(
            caption=caption,
            num_total=len(crunched_entries),
            duration_slowest=slowest,
            duration_fastest=fastest,
            first_sent=first_sent,
            last_received=last_received,
            total_duration=total_duration,
            total_aggregated_time=total_aggregated_time,
            duration_avg=avg,
        )


class GlobalStatistic:
    def __init__(self) -> None:
        self._created_at = time.time()
        self._entries: List[Total] = []

    def __len__(self) -> int:
        return len(self._entries)

    def add(self, total: Total) -> None:
        self._entries.append(total)
